/* Pdf417decode.c

Pdf417decode can decode PDF417 barcode information from a pbm file. 
The decoder is rather simple, and ignores error correction info.
Error correction for pdf417 is Reed Solomon error correction.
You can see Reed Solomon error correction information at
http://www.geocities.com/ooosawaddee3pdf417/


Usage:To decode a pbm file "jac.pbm", do "./pdf417decode jac.pbm".  
The file is written to stdout.

The decoder can take a -d option to output the codewords it saw 
in the image.

Notes:

The decoder ignores error correction info.
Your compiler needs to understand that "long long" is 64 bits (gcc does) 
in order to compile it.

Thank for basic original software by Ian Goldberg, 
Modify and update by OOO S.  (ooosawaddee3@hotmail.com)

If you can debug or modify this software to work better than now 
(like add error correction or modify for any operation system) 
you can cantact OOO S.  (ooosawaddee3@hotmail.com)
I will annouce your name in the next modify rev.

For specification of pdf417 you can see at http://www.geocities.com/ooosawaddee3pdf417/ 
*/										 
/*           
			 Modify and Update
			 22/12/01         pdf417decode   
			 Initial Release 
							  
			 23/12/01         pdf417decode rev 1.1	      
			 Add a test to see if fopen() suceeded.  */


#include <stdio.h>
#include <pbm.h>

/* You may have to play with these numbers, depending on your scan quality */
#define FUZZ_THRESH (cols/40)
#define ROW_THRESH (cols/20)

typedef unsigned int UInt32;
typedef int Int32;
/* You'd better be using gcc, or else an Alpha, or something */
/*Your compiler needs to understand that "long long" is 64 bits (gcc does) 
in order to compile it.*/ 

typedef unsigned long long UInt64;

extern const unsigned dham[3][32768];

int mask[15];

int debug = 0;

/* Return -1 if wrong cluster */
static int bestham(int word, int which) {
    int i, best = which;
    if (debug) printf("%d %.4x %.5x\n", which, word, dham[which][word]);
    for(i=0;i<3;++i) {
	if ((dham[i][word] & 0xff000000) < (dham[best][word] & 0xff000000)) {
	    best = i;
	}
    }
    if (best != which) return -1;
    return dham[which][word];
}

static void outit(int word) {
    static numouts = 0;
    static len = 0;
    static type = 0;
    static sorow = 0;
    static Int32 codewords[34*90];
    static skip = 0;
    if (skip) {
	--skip;
	return;
    }
    if (numouts == 0) {
	len = word;
    }
    if (numouts == 1) {
	type = word;
    }
    if ((word & 0xffffff) == 0x030000) {
	skip = 1;
	sorow = numouts;
	return;
    }
    if ((word & 0xffffff) == 0x030001) {
	--numouts;
	return;
    }
    if (word == -2) {
	int i;
	for(i=2;i<len-4;i+=5) {
	    /* Convert from base 900 to base 256 */
	    int j;
	    unsigned char b[6];
	    UInt64 codeval = 0;

	    if (i==len-5 && type!=924) break;
	    for(j=0;j<5;++j) {
		codeval *= 900;
		codeval += codewords[i+j];
	    }
	    for(j=0;j<6;++j) {
		b[5-j] = codeval%256;
		codeval >>= 8;
	    }
	    fwrite(b, 6, 1, stdout);
	}
	for(;i<len;++i) {
	    unsigned char b = codewords[i];
	    fwrite(&b, 1, 1, stdout);
	}
	fflush(stdout);
    }
    if (word == -3) {
	/* Rewind to beginning of row */
	numouts = sorow;
	return;
    }
    codewords[numouts++] = word & 0xffff;
}

int processrow(int cols, int rownum, int num, double *cumbits)
{
    int firstblack = 0;
    int scale;
    int j;
    int nchange;
    int cumchange[cols];
    double thresh = 0.5*num;
    while(firstblack < cols && cumbits[firstblack] < thresh)
	++firstblack;
    if (firstblack+1 >= cols) return 0;
    nchange = 0;
    cumchange[nchange++] = 0;
    for(j=firstblack+1;j<cols;++j) {
	if ((cumbits[j]<thresh) != (cumbits[j-1]<thresh)) {
	    if (nchange>1 &&
		(j-firstblack-cumchange[nchange-1])*15<cumchange[1]) {
		/* Spurious change */
		--nchange;
	    } else {
		cumchange[nchange++] = j-firstblack;
	    }
	}
    }
#if 0
    for(j=0;j<nchange;++j) {
	printf("%3d ", cumchange[j]);
    }
    printf("\n");
#endif
    if (nchange < 8) return 0;
    for(j=0;j<nchange-8;j+=8) {
	int k;
	int word = 0;
	unsigned cw;
	scale = cumchange[j+8]-cumchange[j];
	for (k=0;k<8;k+=2) {
	    int s,e,l;
	    s = (int)(17.0*(double)(cumchange[j+k]-cumchange[j])/(double)scale+0.5);
	    e = (int)(17.0*(double)(cumchange[j+k+1]-cumchange[j])/(double)scale+0.5);
	    /* We know we always start with 1, end with 0 */
	    if (s < 1) s = 1;
	    if (e > 16) e = 16;
	    for(l=s;l<e;++l) word |= mask[l-1];
	}
	word >>= 1;
	cw = bestham(word, rownum%3);
	if (cw == -1) {
	    outit(-3);
	    return 0;
	}
	outit(cw);
    }
    return 1;
}

int main(int argc, char **argv)
{
    FILE *pf;
    int rows, cols;
    bit **bits;
    double *cumbits;
    int i,j;
    int ready,num;
    int rownum;

    for (i=0;i<15;++i) mask[i] = 1<<(15-i);
    
    if (argc > 1 && !strcmp(argv[1], "-d")) {
	debug = 1;
	--argc;
	++argv;
    }
    pbm_init(&argc, argv);
    
/*Add a test to see if fopen() suceeded. */
     
    if (argc > 1)
      pf = fopen(argv[1], "r");
    else {
      fprintf(stderr, "usage: pdf417decode [-d] file\n");
      exit(1);
    }

    if (pf == NULL) {
      fprintf(stderr, "could not open file: %s\n"), argv[1];
      exit(1);
    }
 
     
    bits = pbm_readpbm(pf, &cols, &rows);
    fclose(pf);
    cumbits = malloc(cols*sizeof(double));

    ready = 1;
    num = 0;
    rownum = 0;
    for(j=0;j<cols;++j) cumbits[j]=0.0;
    for(i=1;i<rows;++i) {
	int d = 0;
	for(j=0;j<cols;++j) {
	    if (bits[i][j] != bits[i-1][j]) ++d;
	}
	if (d < FUZZ_THRESH) {
	    if (ready == 1) {
		num = 0;
		for(j=0;j<cols;++j) cumbits[j] = 0.0;
		ready = 2;
	    }
	    for(j=0;j<cols;++j)
		if (bits[i][j] == PBM_BLACK) cumbits[j] += 1.0;
	    ++num;
	} else if (d > ROW_THRESH) {
	    if (ready == 2) {
		if (processrow(cols, rownum, num, cumbits)) {
		    ++rownum;
		}
		ready=1;
	    }
	}
    }
    if (ready == 2) if (processrow(cols, rownum, num, cumbits)) ++rownum;
    outit(-2);

    pbm_freearray(bits, rows);
    return 0;
}
